/*
 * @(#)Configuration.java	1.00 11/30/00
 * Copyright (c) 2000 by Paul Regenhardt,  All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Paul Regenhardt ("Confidential Information"). You shall not 
 * disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Paul Regenhardt.
 */
import java.io.*;
import java.net.*;
import java.util.*;

/**
 * This class handles network elements by acting as a
 * Telnet client to the network element using RFC 854.
 * 
 * One object is instantiated for each client session and 
 * monitored in the thread.  The constructor autostarts the 
 * thread.
 * 
 * @author Paul Regenhardt
 * @version 1.0
 */


public class TelnetThread extends Thread
{

    /**
     * The socket connection to the destination
     * device
     */
    private Socket aSocket;

    /**
     * The input stream from the socket
     */
    private BufferedReader input;

    /**
     * The output steam to the socket
     */
    private PrintStream output;

    /**
     * The owner object of this handler
     */
    private TelnetDeamon owner;

    /**
     * The strings buffered to be sent
     */
    private Vector buffer = new Vector();
    
    /**
     * The terminator to process input
     */
    private int TERMINATOR = ';'; // for TL1
    //private int TERMINATOR = 13;  // For CR line terminator
    
    /**
     * A flag to abort output
     */
    private boolean outputFlag = true;

    /**
     * A flag to indicate if the input should be echoed back out
     */
    private boolean echo = true;

    /**
     * A flag to indicate the connection/user was validated
     */
    private boolean authenticated = false;

    /**
     * The input prompt
     */
    private String prompt = new String();

    /**
     * A string of characters that represent the input stream.
     */
    private char action[] = new char[255];

    /**
     * an index into the input stream
     */
    private int  index = 0;

    /**
     * Number of minutes to time out the connection
     */
    private int  timeOut = 0; // infinite timeout = 0; otherwise minutes

    /**
     * The last time a command was executed
     */
    private long  lastCommandTime;

    private char IAC    = (char) 255;
    private char WILL   = (char) 251;
    private char WONT   = (char) 252;
    private char DONT   = (char) 254;
    private char GA     = (char) 249;
    private char EL     = (char) 248;
    private char EC     = (char) 247;
    private char AYT    = (char) 246;
    private char SB     = (char) 250; // subnegotiate
    private char SE     = (char) 240; // subnegotiate end
    private char NOP    = (char) 241;
    private char DMK    = (char) 242;
    private char BREAK  = (char) 243;
    private char IP     = (char) 244; // Interupt Process
    private char AO     = (char) 245; // Abort Output
    private char TT     = (char) 24; // Terminal Type
    private char CR     = (char) 13; // Carriage Return
    private char LF     = (char) 10; // Line Feed

    /**
     * Shortcut method to send a carriage-control and linefeed to the telnet client
     */
    public void sendCRLF() {output.print(CR); output.print(LF);}

    /**
     * Shortcut method to send a carriage return to the telnet client
     */
    public void sendCR() {output.print(CR); }

    /**
     * Shortcut methode to send a line feed to the telnet client
     */
    public void sendLF() {output.print(LF); }

    /**
     * Sets the timeout (in minutes) for the telnet client.  If the client has not entered any new data 
     * within the timeout period then the connection is closed.
     * 
     * @param min
     */
    public void setTimeout(int min) {timeOut = min; }
    
	/**
	 * Instantiates the object, assigns the user, 
	 * establishes the connection, and starts the 
	 * thread.
	 * 
	 * @param o The user object
	 * @param s The socket
	 */
	public TelnetThread(TelnetDeamon o, Socket s)
	{
	    aSocket = s;
	    owner = o;
	    try {
	       output = new PrintStream(aSocket.getOutputStream());
           input = new BufferedReader(new InputStreamReader(aSocket.getInputStream()));
        } catch (Exception e) {System.out.println(e);}
        start();
	}

    /**
     * Telnet connectivity test
     */
    public void areYouThere() 
    {
	    output.print(IAC);
        output.print(AYT);
    }
    
    /**
     * sets the authentication flag.
     * 
     * @param flag
     */
    public void authenticate(boolean flag)
    {
        authenticated = flag;
    }
    
    /**
     * responds true if the user has been authenticated
     */
    public boolean isAuthenticated() {return authenticated;}

    /**
     * Telnet Interuption service
     */
    public void interupt() 
    {
	    output.print(IAC);
        output.print(IP);
	    output.print(IAC);
        output.print(AO);
    }

    /**
     * This method transmits one line at
     * a time to the output destination.
     * 
     * It appends a CR/LF and waits for the
     * data to be transmitted via a sleep
     * before allowing the next line.
     * 
     * @param s The string to output
     */
    public void transmit(String s) 
    {
	    if (!outputFlag) return; // abort output?
	    if (s == null) return;
	    if (s.length() < 1) return;
	    
	    output.print(s);
        try {sleep(100);} catch (Exception e) {}
    }
	
	/**
	 * Final process cleanup
	 */
	public void cleanup()
	{
	    try {
    	    output.close();
	        input.close();
	        aSocket.close();
        } catch (Exception e) {}
	}
	
	/**
	 * set's the prompt that the telnet client will see when waiting for input
	 * 
	 * @param p
	 */
	public void setPrompt(String p)
	{
	    prompt = p;
	}
	
    /**
     * Gets a string of input from the telnet client terminated by the termination character
     */
    public String getString()
    {
        sendCRLF();
        transmit(prompt);
        if (timeOut != 0) lastCommandTime = (new Date()).getTime();
        
        index = 0;
        while (true) {
	        try {
    	        while (input.ready()) {
	                int i = input.read();
		    	    if (i == IAC) processIAC();
        	        else {
    	                if (i == TERMINATOR) { // Answer the action at the end
    	                    action[index] = 0; // append a null at the end of the action
                            if (echo) output.print((char) i);
    	                    if (echo) sendCRLF();
    	                    return new String(action, 0, index);
                        } else if (i == 7 || (i >= 32 && i < 127) ) { // append only real characters 
                            if (echo) output.print((char) i);
			                action[index++] = (char) i;
			            }
			        }
			    }
			    try {sleep(100); } catch (InterruptedException e) {
			        System.out.println(e);
			        }
	            yield(); // yield if there is no input waiting
	            if (timeOut != 0) {
	                long timeDelta = (new Date()).getTime() - lastCommandTime;
	                if (timeDelta/1000/60 > timeOut) {
	                    transmit("\nSystem timed out, connection lost\n");
	                    owner.closeConnection(this);  // no return, thread stopped
	                }
	            }
		    } catch (Exception e) {
		        System.out.println(e);}
         }
	}

	/**
	 * Processes the Telnet IAC, essentially
	 * it refuses everything except 
	 * terminal type which is set to unknown (not needed for a server).
	 * 
	 * @exception java.io.IOException
	 */
	private void processIAC() throws java.io.IOException
	{
	    int j;
        int i = input.read();
    	if (i==253) { // DOs
	        j = input.read();
    	    switch(j) {
    	    case 24: // request terminal type
                 output.print(IAC); 
    	         output.print(WILL); 
    	         output.print((char) j);
    	         break;
    	    default: // refuse any unknown requests
        	     output.print(IAC); 
    	         output.print(WONT); 
    	         output.print((char) j);
    	    }
    	} else if (i == WILL) {// WILL -- skip
	        j = input.read();
    	} else if (i == WONT) {// WONT -- skip
	        j = input.read();
    	} else if (i == DONT) {// DONT -- skip
	        j = input.read();
    	} else if (i == 250) {// subnegotiate
	        j = input.read();
     	    switch(j) {
    	    case 24: // request terminal type
     	         //Server: IAC SB TERMINAL-TYPE SEND IAC SE
                 //Client: IAC SB TERMINAL-TYPE IS UNKNOWN IAC SE
	             j = input.read();
	             if (j == 1) { // "send" requested to send the TT
        	        output.print(IAC);
    	            output.print(SB);
    	            output.print((char) TT);  //TERMINAL-TYPE
    	            output.print((char)0);   // is
    	            output.print("UNKNOWN");
        	        output.print(IAC); 
    	            output.print(SE);
    	         }
    	         // skip to the end of the SB
    	         while (j != SE) j=input.read(); // we ignore the terminal type here, in a real implementation you'd want to handle this
    	         
    	         break;
    	    default: // refuse
    	         input.skip(3);
    	    }
    	} else if (i == AO)
    	    outputFlag = false;
    	else if (i == EC)
    	    if (index > 0) index--;
    	else if (i == EL)
    	    index = 0;
	}
	
	/**
	 * Some telnet clients expect this to be asked, not demanded by the RFC.
	 */
	public void setupLink()
	{ 
	    //Server: IAC SB TERMINAL-TYPE SEND IAC SE
        output.print(IAC); 
    	output.print(SB); 
    	output.print(TT);
    	output.print((char)1);   // SEND
        output.print(IAC); 
        output.print(SE); 
	}
	
	/**
	 * This is the main thread loop to 
	 * process data
	 */
	public void run()
	{
	    String recvd;
	    
	    while (true)  {
	        recvd = getString(); // won't return until there is data
            outputFlag = true; // user requested output so turn it on
	        owner.performAction(this, recvd); // process input from user
	        yield(); // yield if there is no input waiting
	    }
	}
}
